#include <assert.h>
#include <assert.h>
#include <QSqlTableModel>
#include <QSqlRecord>
#include <QDebug>
#include <QSqlError>
#include <QSqlQuery>
#include <QFuture>
#include <QtConcurrentRun>
#include "creaturemanager.h"
#include "zonemanager.h"
#include "fieldmanager.h"
#include "marketmanager.h"
#include "stockmanager.h"
#include "speciesmanager.h"
#include "recipeexecutor.h"
#include "util/random.h"

void CreatureManager::executeActivities(Creature &creature)
{
    // Określamy pole z jakim związany jest osobnik
    FieldManager *fieldManager = FieldManager::getManager();
    MarketManager *marketManager = MarketManager::getManager();
    StockManager *stockManager = StockManager::getManager();
    const Field *creatureField = fieldManager->objectAtPosition(creature.getCoordinates().first,
                                                                creature.getCoordinates().second);
    const Market *market = marketManager->objectAtPosition(creature.getCoordinates().first,
                                                           creature.getCoordinates().second);

    // Przygotujmy zestaw zapasów
    Stock::Set containers;
    if (creatureField != NULL) {
        // Kopiujemy pole i sklep pola
        stockManager->appendTo (containers, Container::FIELD, creatureField->getResourcesId());

    }
    if (market != NULL) {
         stockManager->appendTo (containers, Container::MARKET, market->getStockId());
    }
    // Oraz zasoby osobnika
    stockManager->appendTo (containers, Container::CREATURE, creature.getBackpackId());

    Creature::Activities::const_iterator actIter = creature.getActivities().begin();
    Creature::Activities::const_iterator actEnd = creature.getActivities().end();
    RecipeExecutor *recipeExecutor = RecipeExecutor::getManager();

    while (actIter != actEnd) {
        QString recipeId = *actIter;
        const Recipe *recipe = recipeExecutor->find(recipeId);
        if (recipe == NULL) {
            qWarning() << "Not found recipe " << recipeId << " for creature " << creature.getId();
        }
        assert (recipe != NULL);

        bool executed = recipeExecutor->execute(*recipe, containers, creature);
        if (messenger != NULL ) {
            messenger->creatureActivityDoneSignal(creature.getId(),
                                                  executed ? recipeId : QString::null);
        }

        actIter++;
    }
    containers.clear();
}

void CreatureManager::dayActivity(QString creatureId)
{
    Cache::iterator f = cache.find(creatureId);
    if (f != cache.end()) {
        Creature *creature = &f->second;
        executeActivities(*creature);
        creature->increaseAge();
    }
}

void CreatureManager::mating(const Creature &mom, const Creature &dad, Creature &child)
{
    SpeciesManager *speciesManager = SpeciesManager::getManager();
    RecipeExecutor *recipeExecutor = RecipeExecutor::getManager();
    StockManager *stockManager = StockManager::getManager();

    const Species *species = speciesManager->find(mom.getSpeciesId());
    assert (species != NULL);
    assert (species == speciesManager->find(dad.getSpeciesId()));

    QString creatureBackpackId = QString ("backpack of %1").arg(child.getId());
    stockManager->copyStock (species->getInitialStockId(),
                             creatureBackpackId);

    child.setBackpackId(creatureBackpackId);
    Coordinates creatureCoordinates( (mom.getCoordinates().first + dad.getCoordinates().first) / 2,
                                     (mom.getCoordinates().second + dad.getCoordinates().second) / 2);
    child.setCoordinates (creatureCoordinates);
    child.setSpecies (*species);
    child.clearActivities();

    Creature::Activities::const_iterator momActIter = mom.getActivities().begin();
    Creature::Activities::const_iterator actEnd = mom.getActivities().end();
    Creature::Activities::const_iterator dadActIter = dad.getActivities().begin();
    // One point crossover
    int genesCount = species->getActivitiesCount();
    int breakGeneIndex = Random::between(0, genesCount);
    bool doCrossover = Random::fromRange (0.0, 1.0) <= species->getCrossoverProbability();
    if (!doCrossover)
        // Jeżeli nie będzie krzyżowania, to wszystkie geny matki
        breakGeneIndex = genesCount;
    int i = 0;
    if (species->getCrossoverType() == "one_point") {
        while (momActIter != actEnd ) {
            const Recipe *recipe = NULL;
            QString parentRecipeId;
            if (i < breakGeneIndex) {
                parentRecipeId = *momActIter;
            } else {
                parentRecipeId = *dadActIter;
            }
            if (Random::fromRange (0.0, 1.0) <= species->getMutationProbability()) {
                const Recipe *recipe = recipeExecutor->randomRecipe(Container::CREATURE);
                if (recipe != NULL)
                    parentRecipeId = recipe->getId();
            }
            recipe = RecipeExecutor::getManager()->find(parentRecipeId);
            child.addActivity(*recipe);
            momActIter++;
            dadActIter++;
            i++;
        }
    } else {
        while (momActIter != actEnd ) {
            const Recipe *recipe = NULL;
            QString parentRecipeId;
            if (Random::flipCoin()) {
                parentRecipeId = *momActIter;
            } else {
                parentRecipeId = *dadActIter;
            }
            if (Random::fromRange (0.0, 1.0) <= species->getMutationProbability()) {
                const Recipe *recipe = recipeExecutor->randomRecipe(Container::CREATURE);
                if (recipe != NULL)
                    parentRecipeId = recipe->getId();
            }
            recipe = RecipeExecutor::getManager()->find(parentRecipeId);
            child.addActivity(*recipe);
            momActIter++;
            dadActIter++;
            i++;
        }
    }
}

const CreatureManager::IdsOfObjects CreatureManager::creaturesInArea(const Area &area,
                                                                     const Species &species) const
{
    IdsOfObjects idsOfCreatures;

    Cache::const_iterator c = cache.begin();
    Cache::const_iterator end = cache.end();
    while (c != end) {
        const Creature *creature = &c->second;
        if (creature->getSpeciesId() == species.getId()
                && area.contains (creature->getCoordinates())) {
            idsOfCreatures.push_back(creature->getId());
        }
        c++;
    }
    return idsOfCreatures;
}

void CreatureManager::reproduction(const Area &area, const Species &species)
{
    IdsOfObjects parents = creaturesInArea(area, species);
    // Z potencjalych rodziców wybieramy losowo dwóch, którzy wydadzą potomstwo
    int parentPoolSize = parents.size();
    if (parentPoolSize > 1) {
        int momIndex = Random::between(0, parentPoolSize -1);
        int dadIndex = Random::between(0, parentPoolSize -1);
        const Creature *mom = find(parents.at(momIndex));
        const Creature *dad = find(parents.at(dadIndex));
        QString childId ("%1 of %2");
        childId = childId.arg(nextObjectId).arg(species.getId());
        Creature child (childId, mom->getImageId());
        // Rodzicę dają dziecie
        mating(*mom, *dad, child);
        // Osobnik trafia do zbioru wszystkich osobników
        merge(child);
        if (messenger != NULL)
            messenger->creatureBornSignal (childId);
    }
}

double CreatureManager::objectiveValue(const Creature &creature) const
{
    //return creature.getBackpackObjectiveValue();
    StockManager *stockManager = StockManager::getManager();
    return stockManager->valueOfStock(creature.getBackpackId());
}

// Używane sortowania osobników względem wartości funkcji celu
typedef std::pair<QString, double> QStringFloatPair;
struct QStringFloatPairCmp {
    bool operator()(const QStringFloatPair &lhs, const QStringFloatPair &rhs) {
        return lhs.second < rhs.second;
    }
};

void CreatureManager::evolutionaryPressure(const Area &area, const Species &species, int creatureLimit)
{
    IdsOfObjects idsOfCreatures = creaturesInArea(area, species);
    int removeCount = idsOfCreatures.size() - creatureLimit;
    if (removeCount > 0) {
        // Trzeba zmniejszyć populację, więc sortujemy względem wartości funkcji celu
        std::vector< QStringFloatPair > sortVector;
        IdsOfObjects::const_iterator c = idsOfCreatures.begin();
        IdsOfObjects::const_iterator end = idsOfCreatures.end();
        while (c != end) {
            const Creature *creature = find(*c);
            double objVal = objectiveValue(*creature);
            QStringFloatPair elem (creature->getId(), objVal);
            sortVector.push_back(elem);
            c++;
        }
        std::sort(sortVector.begin(), sortVector.end(), QStringFloatPairCmp());
        // Usuwamy pierwsze z brzegu
        do  {
            QString creatureId = sortVector.begin()->first;
            if (messenger)
                messenger->creatureDiedSignal(creatureId);
            const Creature *creature = find(creatureId);
            // Zostawiamy, to co zebraliśmy w ciągu życia na polu
            Coordinates coord = creature->getCoordinates();
            const Field *field = FieldManager::getManager()->objectAtPosition(coord.x(), coord.y());
            if (field) {
                StockManager *stockManager = StockManager::getManager();
                // TODO - tylko te artykuły, które można zostawić przerzucamy do zasobów
                // pola.
                const Stock *stock = stockManager->find(field->getResourcesId());
                if (stock) {
                    stockManager->moveContent(creature->getBackpackId(),
                                              field->getResourcesId());
                } else {
                    qWarning() << "Brak zasobów " << field->getResourcesId() << " osobnika " << field->getId();
                }
            }
            remove(creatureId);
            removeCount--;
            sortVector.erase(sortVector.begin());
            ;
        }
        while (removeCount > 0 && sortVector.size()) ;
    }
}


bool CreatureManager::load()
{
    StockManager *stockManager = StockManager::getManager();
    SpeciesManager *speciesManager = SpeciesManager::getManager();
    RecipeExecutor *recipeExecutor = RecipeExecutor::getManager();
    ZoneManager *zoneManager = ZoneManager::getManager();
    QSqlQueryModel model;
    model.setQuery("SELECT * FROM creature");
    for (int i = 0; i < model.rowCount(); ++i) {
        QSqlRecord r = model.record(i);
        QString id = r.value("id").toString();
        QString speciesId = r.value("species_id").toString();
        QString backpackId = r.value("backpack_id").toString();
        Coordinates coordinates (r.value("position_x").toInt(),
                                 r.value("position_y").toInt());
        int age = r.value("age").toInt();
        QString imageId = r.value("image_id").toString();

        const Stock *backpack = stockManager->find(backpackId);
        const Species *species = speciesManager->find(speciesId);

        if (backpack == NULL)
            qFatal("%s", QString ("backpack %1 not found while loading creature %2").arg(backpackId).arg(id).toStdString().c_str());
        if (species == NULL)
            qFatal("%s", QString ("species %1 not found while loading creature %2").arg(speciesId).arg(id).toStdString().c_str() );

        Creature creature (id, coordinates, *species, *backpack, age, imageId);
        creature.move(0, 0, zoneManager->getWorldArea());

        // Dla danego osobnika wczytujemy jego aktywności
        QSqlQueryModel activities;
        activities.setQuery(QString("SELECT * FROM activity WHERE creature_id = '%1' ORDER BY LP").arg(id));
        for (int j = 0; j < activities.rowCount(); ++j) {
            QSqlRecord v = activities.record(j);
            QString activityId = v.value("definition_id").toString();
            // int lp = v.value("lp").toInt();
            const Recipe *recipe = recipeExecutor->find(activityId);
            if (recipe == NULL)
                qFatal("%s", QString ("recipe %1 not found while loading creature %2").arg(activityId).arg(id).toStdString().c_str() );
            creature.addActivity(*recipe);
            qDebug() << "activity " << activityId << " of creature" << id << "loaded";
        }
        merge(creature);
        qDebug() << "creature " << creature.getId() << " loaded";
    }
    return true;
}

bool CreatureManager::save(const Creature &creature)
{
    QMap<QString, QVariant> values;
    values.insert ("id", creature.getId());
    values.insert ("species_id", creature.getSpeciesId());
    values.insert ("age", creature.getAge());
    values.insert ("backpack_id", creature.getBackpackId());
    values.insert ("position_x", creature.getCoordinates().x());
    values.insert ("position_y", creature.getCoordinates().y());
    values.insert ("image_id", creature.getImageId());
    bool saved = SingletonManagerTemplate::save (values, creature.getId());
    Creature::Activities::const_iterator ab = creature.getActivities().begin();
    Creature::Activities::const_iterator ae = creature.getActivities().end();
    saved = removeRows ("activity",  QString ("creature_id = '%1'").
                        arg(creature.getId()));
    int lp = 1;
    while (saved && ab != ae) {
        values.clear();
        values.insert("creature_id", creature.getId());
        values.insert("lp", lp);
        values.insert("definition_id", *ab);
        saved = SingletonManagerTemplate::save ("activity",
                                                QString ("creature_id = '%1' and lp = %2").
                                                arg(creature.getId()).arg(lp),
                                                values);
        ab++;
        lp++;
    }
    return saved;
}

void CreatureManager::setMessenger(Messenger *emitter)
{
    this->messenger = emitter;
}

void CreatureManager::init()
{
    SingletonManagerTemplate::init();
    messenger = NULL;
    nextObjectId = 1;
}

void CreatureManager::merge(const Creature &object)
{
    SingletonManagerTemplate::merge(object);
    nextObjectId++;
}

// Usuń z cache
void CreatureManager::remove(QString id) {
    StockManager::getManager()->remove(find(id)->getBackpackId());
    SingletonManagerTemplate::remove(id);
}


void CreatureManager::setNextObjectId(unsigned long next)
{
    nextObjectId = next;
}

unsigned long CreatureManager::getNextObjectId () {
    return nextObjectId;
}

bool CreatureManager::removeDependent(QString creatureId)
{
    return SingletonManagerTemplate::removeDependent (creatureId)
            && removeRows ("activity", QString("creature_id = '%1'").arg(creatureId));
}
