#include <assert.h>
#include <list>
#include <QSqlTableModel>
#include <QSqlRecord>
#include <QDebug>
#include <QSqlError>
#include <QSqlQuery>

#include "recipeexecutor.h"
#include "articlemanager.h"
#include "zonemanager.h"
#include "util/random.h"

bool RecipeExecutor::execute(const Recipe &recipe,
                             Stock::Set &stocks)
{
    // Stwórz produkty w odpowiednich ilościach i wypełnij docelowe kontenery
    bool executed = true;

    std::vector<StockLock> containers;
    // Zablokuj dostęp do kontenerów i wypełnij containers pojemnikami, które faktycznie będziemy używać
    executed = necessaryStocksLocked (recipe, stocks, containers);

    // Mamy zablokowane wszystkie potrzebne zasoby, możemy "warzyć" recepturę.
    executed = executed
            && haveAvailableQuants (recipe, stocks);

    // Sprawdź dostępność miejsca
    executed = executed
            && containersCapacious(recipe, stocks);


    if (executed) {
        // Pobieramy ilości z zasobów
        Recipe::Resources::const_iterator resourceIter = recipe.getResources().begin();
        Recipe::Resources::const_iterator lastResource = recipe.getResources().end();
        for (; resourceIter != lastResource; resourceIter++) {
            Recipe::Element resource = *resourceIter;
            Stock *container = stocks.at(resource.container);
            const Article *resourceArticle = ArticleManager::getManager()->find(resource.articleId);
            container->pop(*resourceArticle, resource.quant);
            if (messenger) {
                messenger->stockChangedSignal(container->getId());
                messenger->articleQuantChangeSignal(container->getId(), resourceArticle->getId(), -resource.quant);
            }
        }

        // Aktualizujemy ilości powstałych produktów
        Recipe::Products::const_iterator productIter = recipe.getProducts().begin();
        Recipe::Products::const_iterator lastProduct = recipe.getProducts().end();
        for (; productIter != lastProduct; productIter++) {
            Recipe::Element product = *productIter;
            Stock *container = stocks.at(product.container);
            const Article *productArticle = ArticleManager::getManager()->find(product.articleId);
            container->push(*productArticle, product.quant);
            if (messenger) {
                messenger->stockChangedSignal(container->getId());
                messenger->articleQuantChangeSignal(container->getId(), productArticle->getId(), product.quant);
            }
        }
    }

    // Odblokuj zablokowane stoki
    std::vector<StockLock>::iterator s = containers.begin();
    std::vector<StockLock>::iterator e = containers.end();
    for (; s != e; s++) {
        if (s->stock != NULL) {
            s->mutex->unlock();
            delete s->mutex;
        }
    }

    return executed;
}

bool RecipeExecutor::execute(const Recipe &recipe,
                             Stock::Set &stocks,
                             Creature &movable)
{
    ZoneManager *fieldManager = ZoneManager::getManager();
    bool executed = RecipeExecutor::execute(recipe, stocks);
    if (executed) {
        // Jeżeli receptura wpływa na położenie inicjatora, to zmień jego położenie
        if ( recipe.getPositionEffect().isEffective() ) {
            Area world = fieldManager->getWorldArea();
            movable.move(recipe.getPositionEffect().deltaX,
                         recipe.getPositionEffect().deltaY,
                         world);

        }

    }
    return executed;
}

const Recipe *RecipeExecutor::randomRecipe(Container::Type owner) const
{
   IdsOfObjects recipesIds = recipesForOwner(owner);
   int vectSize = recipesIds.size();
   if (vectSize > 0) {
       int index = Random::between(0, vectSize - 1);
       return find(recipesIds.at(index));
   }
   return NULL;
}

const RecipeExecutor::IdsOfObjects RecipeExecutor::recipesForOwner(Container::Type owner) const
{
    IdsOfObjects idsOfRecipes;
   // idsOfCreatures.assign(0,"");
    Cache::const_iterator r = cache.begin();
    Cache::const_iterator end = cache.end();
    while (r != end) {
        const Recipe *recipe = &r->second;
        if (recipe->getOwner() == owner) {
            idsOfRecipes.push_back(recipe->getId());
        }
        r++;
    }
    return idsOfRecipes;
}

bool RecipeExecutor::necessaryStocksLocked(const Recipe &recipe,
                                           Stock::Set &stocks,
                                           std::vector<StockLock> &necessaryStocks)
{
    bool locked = true;
    // Pobierz na wyłączność kontenery, które będą brały udział w recepturze
    Recipe::ContainerTypes::const_iterator recipeContainersIter = recipe.getRequiredContainerTypes().begin();
    Recipe::ContainerTypes::const_iterator recipeContainersEnd = recipe.getRequiredContainerTypes().end();
    // Zablokuj je
    StockLock nullStockLock;
    nullStockLock.mutex = NULL;
    nullStockLock.stock = NULL;

    necessaryStocks.assign(Container::__CONTAINER__COUNT, nullStockLock);
    for (; locked && recipeContainersIter != recipeContainersEnd;
         recipeContainersIter ++) {
        Stock::Set::iterator reqStockIter = stocks.find(*recipeContainersIter);
        locked = reqStockIter != stocks.end();
        if (locked) {            
            StockLock stockLock;
            stockLock.mutex = new QMutex();


            if ( (locked = stockLock.mutex->tryLock()) ) {
                Stock *stock = reqStockIter->second;//stocks.at(*recipeContainersIter);
                stockLock.stock = stock;
                necessaryStocks.at(*recipeContainersIter) = stockLock;
            } else {
                delete stockLock.mutex;
            }
        }
    }
    return locked;
}

bool RecipeExecutor::haveAvailableQuants (const Recipe &recipe, const Stock::Set &stocks) {
    bool have = true;
    Recipe::Resources::const_iterator resourceIter = recipe.getResources().begin();
    Recipe::Resources::const_iterator lastResource = recipe.getResources().end();
    // Sprawdzamy czy mamy wystarczające ilości
    for (; have && resourceIter != lastResource; resourceIter++) {
        //Recipe::Element resource = recipe.getResources() *resourceIter;
        double availableQuant =
                stocks.at(resourceIter->container)->summaryQuant(resourceIter->articleId);
        have = availableQuant >= resourceIter->quant;
    }
    return have;
}

bool RecipeExecutor::containersCapacious(const Recipe &recipe,
                                        const Stock::Set &stocks) {
    bool capacious = true;
    Recipe::Products::const_iterator productIter = recipe.getProducts().begin();
    Recipe::Products::const_iterator lastProduct = recipe.getProducts().end();
    // Sprawdzamy, czy zmieścimy się w docelowych kontenerach
    std::vector<double> volumeSummary;
    volumeSummary.assign(Container::__CONTAINER__COUNT, 0.0);
    for (; capacious && productIter != lastProduct; productIter++) {
        Recipe::Element product = *productIter;
        double articleVolume = ArticleManager::getManager()->find(product.articleId)->volume(product.quant);
        volumeSummary.at(product.container) += articleVolume;
        capacious = volumeSummary.at(product.container) <
                stocks.at(product.container)->availableCapacity();
    }

    return capacious;
}


bool RecipeExecutor::load()
{
    QSqlQueryModel model;
    model.setQuery("SELECT * FROM recipe");
    for (int i = 0; i < model.rowCount(); ++i) {
        QSqlRecord r = model.record(i);
        QString id = r.value("id").toString();
        Container::Type owner = Container::toType (r.value("owner_id").toString());
        int positionXInf = r.value("position_x_influence").toInt();
        int positionYInf = r.value("position_y_influence").toInt();
        Recipe recipe (id, owner, positionXInf, positionYInf);

        // Dla danej receptury wczytujemy jej składniki
        QSqlQueryModel ingredients;
        ingredients.setQuery(QString("SELECT * FROM ingredient WHERE recipe_id = '%1'").arg(id));
        for (int j = 0; j < ingredients.rowCount(); ++j) {
            QSqlRecord v = ingredients.record(j);
         //   int lp = v.value("lp").toInt();
            QString role = v.value("article_role").toString();
            QString articleId = v.value("article_id").toString();
            double quant = v.value("article_quant").toFloat();
            Container::Type container = Container::toType(v.value("article_container").toString());
            if (role.toLower() == "product") {
                recipe.addProduct(container, articleId, quant);
            } else {
                recipe.addResource(container, articleId, quant);
            }
            qDebug() << "ingredient" << articleId << " of recipe " << id << "loaded";
        }
        merge(recipe);
        qDebug() << "recipe" << recipe.getId() << " loaded";
    }
    return true;
}

bool RecipeExecutor::save(const Recipe &recipe)
{
    QMap<QString, QVariant> values;
    values.insert ("id", recipe.getId());
    values.insert ("owner_id", Container::toString(recipe.getOwner()));
    values.insert ("position_x_influence", recipe.getPositionEffect().deltaX);
    values.insert ("position_y_influence", recipe.getPositionEffect().deltaY);

    bool saved = SingletonManagerTemplate::save (values, recipe.getId());
    Recipe::Products::const_iterator pb = recipe.getProducts().begin();
    Recipe::Products::const_iterator pe = recipe.getProducts().end();
    saved = removeRows ("ingredient",  QString ("recipe_id = '%1'").
                        arg(recipe.getId()));

    while (saved && pb != pe) {
        values.clear();
        values.insert("recipe_id", recipe.getId());
        values.insert("article_id", pb->articleId);
        values.insert("article_quant", pb->quant);
        values.insert("article_container", Container::toString(pb->container));
        values.insert("article_role", "product");
        saved = SingletonManagerTemplate::save ("ingredient",
                                                QString ("recipe_id = '%1' and article_id='%2' and "
                                                         "article_container='%3' and article_role='product'").
                                                arg(recipe.getId()).
                                                arg(pb->articleId).
                                                arg(Container::toString(pb->container)),
                                                values);
        pb++;
    }
    Recipe::Resources::const_iterator rb = recipe.getResources().begin();
    Recipe::Resources::const_iterator re = recipe.getResources().end();
    while (saved && rb != re) {
        values.clear();
        values.insert("recipe_id", recipe.getId());
        values.insert("article_id", rb->articleId);
        values.insert("article_quant", rb->quant);
        values.insert("article_container", Container::toString(rb->container));
        values.insert("article_role", "resource");
        saved = SingletonManagerTemplate::save ("ingredient",
                                                QString ("recipe_id = '%1' and article_id='%2' and "
                                                         "article_container='%3' and article_role='resource'").
                                                arg(recipe.getId()).
                                                arg(rb->articleId).
                                                arg(Container::toString(rb->container)),
                                                values);
        rb++;
    }
    return saved;
}

void RecipeExecutor::init()
{
    SingletonManagerTemplate::init();
    messenger = NULL;
}

void RecipeExecutor::setMessenger(Messenger *worldMessenger)
{
    messenger = worldMessenger;
}
