#include <assert.h>
#include <QSqlTableModel>
#include <QSqlRecord>
#include <QDebug>
#include <QSqlError>
#include <QSqlQuery>
#include "stockmanager.h"
#include "articlemanager.h"

StockManager::StockManager()
{
}

void StockManager::appendTo(Stock::Set &containers, Container::Type owner, QString stockId)
{
    std::pair<Container::Type, Stock *> ownerAndItsStock;
    ownerAndItsStock.first = owner;
    Cache::iterator f = cache.find(stockId);
    ownerAndItsStock.second = &(f->second);
    containers.insert(ownerAndItsStock);
}

void StockManager::copyStock(QString sourceId, QString destId)
{
    const Stock *source = find(sourceId);
    if (source != NULL) {
        Stock dest (destId, source->getCapacity());
        ArticleManager *articleManager = ArticleManager::getManager();
        Stock::Articles::const_iterator articleIter = source->getArticles().begin();
        Stock::Articles::const_iterator articlesEnd = source->getArticles().end();
        while (articleIter != articlesEnd) {
            const Article *article = articleManager->find (articleIter->first);
            Stock::ArticleQuants::const_iterator qIter = articleIter->second.begin();
            Stock::ArticleQuants::const_iterator qEnd = articleIter->second.end();
            while (qIter != qEnd) {
                dest.push(*article, qIter->quant, qIter->shelfTime);
                qIter++;
            }
            articleIter++;
        }
        merge(dest);
        if (messenger)  {
            messenger->stockChangedSignal(destId);
            infoAboutArticlesQuantChanges (dest);

        }
    } else {
        qFatal("source stock %s not found when copying", sourceId.toStdString().c_str());
    }
}


double StockManager::valueOfStock(QString stockId)
{
    const Stock *stock = find(stockId);
    if (stock == NULL) {
        qFatal("%s", QString("Stock Id %1 not found").arg(stockId).toStdString().c_str());
    }
    assert(stock != NULL);
    double objValue = 0.0;
    Stock::Articles::const_iterator articleIter = stock->getArticles().begin();
    Stock::Articles::const_iterator articlesEnd = stock->getArticles().end();
    while (articleIter != articlesEnd) {
        const Article *article = ArticleManager::getManager()->find (articleIter->first);
        Stock::ArticleQuants::const_iterator qIter = articleIter->second.begin();
        Stock::ArticleQuants::const_iterator qEnd = articleIter->second.end();
        while (qIter != qEnd) {
            objValue += article->valueForObjectiveFunction(qIter->quant);
            qIter++;
        }
        articleIter++;
    }
    return objValue;
}

void StockManager::dayPassed(QString stockId)
{
    Cache::iterator f = cache.find(stockId);
    Stock *stock = &f->second; //find(stockId);
    assert(stock != NULL);
    bool stockChanged;
    Stock copy = *stock;
    stock->dayPassed(&stockChanged);
    if (stockChanged && messenger) {
        messenger->stockChangedSignal(stockId);

        Stock::Articles::const_iterator articleIter = copy.getArticles().begin();
        Stock::Articles::const_iterator articlesEnd = copy.getArticles().end();
        while (articleIter != articlesEnd) {
            double before = copy.summaryQuant(articleIter->first);
            double after = stock->summaryQuant(articleIter->first);
            if (before != after) {
                messenger->articleQuantChangeSignal(stockId, articleIter->first, after - before);
            }
            articleIter++;
        }
    }
}



bool StockManager::load()
{
    ArticleManager *articleManager = ArticleManager::getManager();
    QSqlQueryModel model;
    model.setQuery("SELECT * FROM stock");
    for (int i = 0; i < model.rowCount(); ++i) {
        QSqlRecord r = model.record(i);
        QString id = r.value("id").toString();
        double capacity = r.value("capacity").toFloat();
        bool shelfTimeChange = r.value("change_shelf_time").toBool();
        Stock stock (id, capacity, shelfTimeChange);
        // Dla danej receptury wczytujemy jej składniki
        QSqlQueryModel articles;
        articles.setQuery(QString("SELECT * FROM inventory WHERE stock_id = '%1'").arg(id));
        for (int j = 0; j < articles.rowCount(); ++j) {
            QSqlRecord v = articles.record(j);
            QString articleId = v.value("article_id").toString();
            int shelfTime = v.value("shelf_time").toInt();
            double quant = v.value("quant").toFloat();
            const Article *article = articleManager->find(articleId);
            if (article == NULL)
                qFatal("%s", QString ("article %1 not found while loading stock %2").arg(articleId).arg(id).toStdString().c_str() );
            stock.push(*article, quant, shelfTime);
            qDebug() << quant << " of article " << articleId << " in stock " << id << "loaded ";
        }
        merge(stock);

        if (messenger) {
            Stock::Articles::const_iterator articleIter = stock.getArticles().begin();
            Stock::Articles::const_iterator articlesEnd = stock.getArticles().end();
            while (articleIter != articlesEnd) {
                double quant = stock.summaryQuant(articleIter->first);
                messenger->articleQuantChangeSignal(id, articleIter->first, quant);
                articleIter++;
            }
        }
        qDebug() << "stock" << stock.getId() << " loaded";
    }
    return true;
}

bool StockManager::save(const Stock &stock)
{
    QMap<QString, QVariant> values;
    values.insert ("id", stock.getId());
    values.insert ("capacity", stock.getCapacity());
    values.insert("change_shelf_time", stock.getChangeShelfTime());
    bool saved = SingletonManagerTemplate::save (values, stock.getId());
    Stock::Articles::const_iterator ab = stock.getArticles().begin();
    Stock::Articles::const_iterator ae = stock.getArticles().end();
    saved = removeRows ("inventory", QString ("stock_id='%1'").arg(stock.getId()));
    while (saved && ab != ae) {
        Stock::ArticleQuants::const_iterator qi = ab->second.begin();
        Stock::ArticleQuants::const_iterator qe = ab->second.end();
        while (qi != qe && saved) {
            values.clear();
            values.insert("stock_id", stock.getId());
            values.insert("article_id", ab->first);
            values.insert("shelf_time", qi->shelfTime);
            values.insert("quant", qi->quant);
            saved = SingletonManagerTemplate::save ("inventory",
                                                    QString ("stock_id = '%1' and article_id='%2' and shelf_time=%3").
                                                    arg(stock.getId()).arg(ab->first).arg(qi->shelfTime), values);

            qi++;
        }
        ab++;
    }
    return saved;
}

void StockManager::remove(QString id)
{
    const Stock *stock = find (id);
    if (stock)
        infoAboutArticlesQuantChanges(*stock, -1);
    else {
        qFatal("%s %s", "Usuwamy coś czego nie możemy znaleźć ??? " , id.toStdString().c_str());
    }
    SingletonManagerTemplate::remove(id);
}

void StockManager::init()
{
    SingletonManagerTemplate::init();
    messenger = NULL;
}

void StockManager::setMessenger(Messenger *worldMessenger)
{
    messenger = worldMessenger;
}

bool StockManager::moveContent(QString sourceStockId, QString destinationStockId, bool partlyMove)
{
    ArticleManager *articleManager = ArticleManager::getManager();
    Cache::iterator sourceIt = cache.find(sourceStockId);
    Cache::iterator destinationIt = cache.find(destinationStockId);
    bool moved = sourceIt != cache.end() && destinationIt != cache.end();
    if (moved) {
        Stock *source = &sourceIt->second;
        Stock *destination = &destinationIt->second;
        // Zapiszemy tutaj co przesuwamy
        Stock movedArticles ("movedArticles", destination->getCapacity());
        moved = partlyMove || source->volumeUsed() <= destination->availableCapacity();
        double quantToMove = 0.;
        if (moved) {
            Stock::Articles::const_iterator saIt = source->getArticles().begin();
            Stock::Articles::const_iterator saEnd = source->getArticles().end();
            while (moved && saIt != saEnd) {
                const Article *article = articleManager->find(saIt->first);
                quantToMove = source->summaryQuant(article->getId());
                if ( destination->availableCapacity() <
                     article->volume(quantToMove) ) {
                    quantToMove = 0.0;
                }

                if (quantToMove > 0.0) {
                    destination->push(*article, quantToMove);
                    movedArticles.push(*article, quantToMove);
                }
                saIt++;
            }

            // Iterujemy po kopii aby nie natrafić na zepsuty iterator
            saIt = movedArticles.getArticles().begin();
            saEnd = movedArticles.getArticles().end();
            while (saIt != saEnd) {
                const Article *article = articleManager->find(saIt->first);
                quantToMove = movedArticles.summaryQuant(article->getId());
                assert(quantToMove > 0);
                source->pop(*article, quantToMove);
                if (messenger) {
                    messenger->articleQuantChangeSignal(destination->getId(), article->getId(), +quantToMove);
                    messenger->articleQuantChangeSignal(source->getId(), article->getId(), -quantToMove);
                }
                saIt++;
            }

            if (moved && messenger) {
                messenger->stockChangedSignal(source->getId());
                messenger->stockChangedSignal(destination->getId());
            }
        }
    } else {
        moved = false;
    }

    return moved;
}

bool StockManager::removeDependent(QString stockId)
{
    return SingletonManagerTemplate::removeDependent (stockId)
            && removeRows("inventory", QString ("stock_id = '%1'").arg(stockId));
}

void StockManager::infoAboutArticlesQuantChanges (const Stock &stock, double sign) {
    Stock::Articles::const_iterator articleIter = stock.getArticles().begin();
    Stock::Articles::const_iterator articlesEnd = stock.getArticles().end();
    while (articleIter != articlesEnd) {
        double q = stock.summaryQuant(articleIter->first);
        messenger->articleQuantChangeSignal(stock.getId(), articleIter->first, sign * q);
        articleIter++;
    }
}
