#include <assert.h>
#include <assert.h>
#include <QSqlTableModel>
#include <QSqlRecord>
#include <QDebug>
#include <QSqlError>
#include <QSqlQuery>
#include "zonemanager.h"
#include "fieldmanager.h"
#include "marketmanager.h"
#include "recipeexecutor.h"
#include "creaturemanager.h"
#include "stockmanager.h"
#include "util/random.h"

Area ZoneManager::worldArea (0,0,0,0);

Zone *ZoneManager::zoneWithPoint(int x, int y)
{
    Zone *zone = NULL;
    Cache::iterator fieldIter = cache.begin();
    Cache::iterator endIter = cache.end();
    while (zone == NULL && fieldIter != endIter) {
        Area fieldArea = fieldIter->second.getArea ();
        if (fieldArea.contains(Coordinates(x, y))) {
            zone = &fieldIter->second;
        }
        fieldIter++;
    }
    return zone;
}

void ZoneManager::merge(const Zone &zone)
{
    SingletonManagerTemplate::merge(zone);
    Area zoneArea = zone.getArea();
    if (cache.size() == 1) {
        worldArea = zoneArea;
    }
    else {
        if (zoneArea.size.first > 0 && zoneArea.size.second > 0) {

            if (worldArea.coordinates.first > zoneArea.coordinates.first) {
                worldArea.size.first += worldArea.coordinates.first - zoneArea.coordinates.first;
                worldArea.coordinates.first = zoneArea.coordinates.first;
            }

            if (worldArea.coordinates.second > zoneArea.coordinates.second) {
                worldArea.size.second += worldArea.coordinates.second - zoneArea.coordinates.second;
                worldArea.coordinates.second = zoneArea.coordinates.second;
            }

            int worldEndX = worldArea.coordinates.first + worldArea.size.first;
            int worldEndY = worldArea.coordinates.second + worldArea.size.second;

            int zoneEndX = zoneArea.coordinates.first + zoneArea.size.first;
            int zoneEndY = zoneArea.coordinates.second + zoneArea.size.second;

            if (worldEndX < zoneEndX)
                worldArea.size.first += zoneEndX - worldEndX;
            if (worldEndY < zoneEndY)
                worldArea.size.second += zoneEndY - worldEndY;
        }
    }
}

const Area &ZoneManager::getWorldArea() const
{
    return worldArea;
}

void ZoneManager::reproduction(const Zone &zone)
{
    // Dla wszystkich gatunków i podobszarów zastosuj reprodukcję
    Zone::LivingConditions::const_iterator lCondIter = zone.getLivingConditions().begin();
    Zone::LivingConditions::const_iterator lCondEnd = zone.getLivingConditions().end();
    CreatureManager *creatureManager = CreatureManager::getManager();
    SpeciesManager  *speciesManager = SpeciesManager::getManager();
    while (lCondIter != lCondEnd) {
        Area zoneArea = zone.getArea();
        const Zone::LivingCondition livingCondition = *lCondIter;
        const Species *species = speciesManager->find(livingCondition.speciesId);
        assert (species != NULL);
        int gridSize = livingCondition.mateGridSize;
        int endX = zoneArea.coordinates.first + zoneArea.size.first;
        for (int x = zoneArea.coordinates.first; x < endX; x += gridSize) {
            int endY = zoneArea.coordinates.second + zoneArea.size.second;
            for (int y = zoneArea.coordinates.second; y < endY; y += gridSize) {
                Area subArea (x, y, gridSize - 1, gridSize -1);
                if (livingCondition.mateProbability >= Random::fromRange(0, 1))
                    creatureManager->reproduction(subArea, *species);
            }
        }
        lCondIter++;
    }
}

void ZoneManager::clear ()
{
    SingletonManagerTemplate::clear();
    worldArea.coordinates.first = 0;
    worldArea.coordinates.second = 0;
    worldArea.size.first = 0;
    worldArea.size.second = 0;
}

void ZoneManager::evolutionaryPressure(const Zone &zone)
{
    // Dla wszystkich gatunków i podobszarów zastosuj ciśnienie ewolucyjne
    Zone::LivingConditions::const_iterator lCondIter = zone.getLivingConditions().begin();
    Zone::LivingConditions::const_iterator lCondEnd = zone.getLivingConditions().end();
    CreatureManager *creatureManager = CreatureManager::getManager();
    SpeciesManager  *speciesManager = SpeciesManager::getManager();
    while (lCondIter != lCondEnd) {
        Area zoneArea = zone.getArea();
        const Zone::LivingCondition livingCondition = *lCondIter;
        const Species *species = speciesManager->find(livingCondition.speciesId);
        assert (species != NULL);
        int gridSize = livingCondition.creaturesLimitGridSize;
        int endX = zoneArea.coordinates.first + zoneArea.size.first;
        for (int x = zoneArea.coordinates.first; x < endX; x += gridSize) {
            int endY = zoneArea.coordinates.second + zoneArea.size.second;
            for (int y = zoneArea.coordinates.second; y < endY; y += gridSize) {
                Area subArea (x, y, gridSize - 1, gridSize -1);
                creatureManager->evolutionaryPressure(subArea, *species, livingCondition.creaturesLimit);
            }
        }
        lCondIter++;
    }
}

bool ZoneManager::load()
{
    SpeciesManager *speciesManager = SpeciesManager::getManager();
    QSqlQueryModel model;

    model.setQuery("SELECT * FROM zone");
    for (int i = 0; i < model.rowCount(); ++i) {
        QSqlRecord r = model.record(i);
        QString id = r.value("id").toString();

        Coordinates coordinates (r.value("coordinate_x").toInt(),
                                 r.value("coordinate_y").toInt());

        QString imageId = r.value("image_id").toString();
        Area zoneArea;
        zoneArea.coordinates = coordinates;
        zoneArea.size.first = r.value("size_x").toInt();
        zoneArea.size.second = r.value("size_y").toInt();

        Zone zone (id, imageId);
        zone.setArea(zoneArea);

        // Dla danego pola wczytujemy jego warunki życia
        QSqlQueryModel livingConditionModel;
        livingConditionModel.setQuery(QString("SELECT * FROM livingcondition WHERE zone_id = '%1' ").arg(id));
        for (int j = 0; j < livingConditionModel.rowCount(); ++j) {
            QSqlRecord v = livingConditionModel.record(j);
            QString speciesId = v.value("species_id").toString();
            int creaturesLimit = v.value("creatures_limit").toInt();
            int creaturesLimitGridSize = v.value("creatures_limit_grid_size").toInt();
            int mateGridSize = v.value("mate_grid_size").toInt();
            double mateProbability = v.value("mate_probability").toDouble();
            const Species *species = speciesManager->find(speciesId);
            if (species == NULL)
                qFatal("%s", QString ("species %1 not found while loading living condition for zone %2").arg(speciesId).arg(id).toStdString().c_str() );

            Zone::LivingCondition livingCondition (*species, creaturesLimit, creaturesLimitGridSize, mateGridSize, mateProbability);
            zone.addLivingCondition(livingCondition);
            qDebug() << "living condition for " << speciesId << " on zone " << id << "loaded";
        }

        merge(zone);
        qDebug() << "zone " << zone.getId() << "(" <<
                    zone.getArea().coordinates.first << "," <<
                    zone.getArea().coordinates.second << "," <<
                    zone.getArea().size.first << "," <<
                    zone.getArea().size.second << ")" <<
                    " loaded";
    }
    return true;
}

bool ZoneManager::save(const Zone &zone)
{
    QMap<QString, QVariant> values;
    values.insert ("id", zone.getId());
    values.insert ("coordinate_x", zone.getArea().x());
    values.insert ("coordinate_y", zone.getArea().y());
    values.insert ("size_x", zone.getArea().width());
    values.insert ("size_y", zone.getArea().height());
    values.insert ("image_id", zone.getImageId());
    bool saved = SingletonManagerTemplate::save (values, zone.getId());

    saved = removeRows ("livingcondition",  QString ("zone_id = '%1'").
                        arg(zone.getId()));
    Zone::LivingConditions::const_iterator li = zone.getLivingConditions().begin();
    Zone::LivingConditions::const_iterator le = zone.getLivingConditions().end();

    while (li != le && saved) {
        values.clear();
        values.insert ("zone_id", zone.getId());
        values.insert ("species_id", li->speciesId);
        values.insert ("creatures_limit", li->creaturesLimit);
        values.insert ("creatures_limit_grid_size", li->creaturesLimitGridSize);
        values.insert ("mate_grid_size", li->mateGridSize);
        values.insert ("mate_probability", li->mateProbability);
        saved = SingletonManagerTemplate::save ("livingcondition",
                                                QString ("zone_id = '%1' and species_id='%2'").
                                                arg(zone.getId()).arg(li->speciesId),
                                                values);
        li++;
    }

    return saved;
}

bool ZoneManager::removeDependent(QString zoneId)
{
    return SingletonManagerTemplate::removeDependent (zoneId)
            && removeRows("livingcondition", QString ("zone_id = '%1'").arg(zoneId))
            ;
}
