#include <assert.h>
#include "stock.h"

Stock::Stock(QString stockId, double startCapacity, bool changeOfShelfTime)
    : Idable<QString> (stockId), articles(), changeShelfTime(changeOfShelfTime)
{
    articlesVolume = 0.0;
    setCapacity(startCapacity);
}

Stock::Stock(const Stock &stock)
    : Idable<QString> (stock.getId())
{
    articles = stock.articles;
    articlesVolume = stock.articlesVolume;
    changeShelfTime = stock.changeShelfTime;
    setCapacity(stock.capacity);
}

Stock::~Stock()
{
    articles.clear();
}


const Stock::Articles &Stock::getArticles() const
{
    return articles;
}

void Stock::push(const Article &article)
{
    push(article, 1.0, article.getShelfTime());
}

void Stock::push(const Article &article, double quant)
{
    push(article, quant, article.getShelfTime());
}

void Stock::push(const Article &article, double quant, unsigned shelfTime)
{
    change (article, quant, shelfTime);
}

void Stock::pop(const Article &article, double quant)
{
    pop(article, quant, -1);
}

void Stock::pop(const Article &article, double quant, unsigned shelfTime)
{
    change (article, -quant, shelfTime);
}

double Stock::summaryQuant(QString articleId, int shelfTime) const
{
    Stock::Articles::const_iterator articleIter = articles.find(articleId);
    double quant = 0.0;
    if ( articleIter != articles.end() ) {
        // Mamy jakieś ilości z tym artykułem,
        const Stock::ArticleQuants *articleQuants = &(articleIter->second);
        // Sumujemy je
        Stock::ArticleQuants::const_iterator first = articleQuants->begin();
        Stock::ArticleQuants::const_iterator last = articleQuants->end();
        while ( first != last ) {
            if (shelfTime <= 0
                    || shelfTime == first->shelfTime)
            quant +=  first->quant;
            first++;
        }
    }
    return quant;
}

double Stock::availableCapacity() const
{
    return capacity - articlesVolume;
}

double Stock::volumeUsed() const
{
    return articlesVolume;
}

int Stock::batchesCount(QString articleId)
{
    Stock::Articles::const_iterator articleIter = articles.find(articleId);
    int count = 0;
    if ( articleIter != articles.end() ) {
        count = articleIter->second.size();
    }
    return count;
}

void Stock::setCapacity(double newCapacity)
{
    assert (articlesVolume <= newCapacity);
    capacity = newCapacity;
}

double Stock::getCapacity() const
{
    return capacity;
}

bool Stock::isEmpty() const
{
    return articles.size() == 0;
}

void Stock::dayPassed(bool *stockChanged)
{
    if (stockChanged) {
        *stockChanged = false;
    }
    if (changeShelfTime) {
        Stock::Articles::iterator articleIter = articles.begin();
        Stock::Articles::iterator articlesEnd = articles.end();
        while (articleIter != articlesEnd) {
            ArticleQuants::iterator qIter = articleIter->second.begin();
            ArticleQuants::iterator qEnd = articleIter->second.end();
            while (qIter != qEnd) {
                qIter->shelfTime --;
                if ( Batch::expireOrNothing(*qIter)) {
                    // Będzie usunięta więc, pojemność trzeba zaktualizować
                    articlesVolume -= qIter->volume;
                    if (stockChanged) {
                        *stockChanged = true;
                    }
                }
                qIter++;
            }
            cleanUp(articleIter);
        }
    }
}

void Stock::setChangeOfShelfTime(bool change)
{
    changeShelfTime = change;
}

bool Stock::getChangeShelfTime() const
{
    return changeShelfTime;
}

void Stock::change (const Article &article, double quant, int shelfTime) {
    QString articleId = article.getId();
    double articleVolume = article.volume(quant);
    articlesVolume += articleVolume;
    assert (articlesVolume <= capacity);
    //assert(shelfTime > 0);
    Stock::Articles::iterator articleIter = articles.find(articleId);
    if ( articleIter == articles.end() ) {
        Stock::Batch batch;//= new Stock::Batch();
        // Nowy wiersz z artykułem
        assert (quant>= 0.0);
        batch.quant = quant;
        batch.shelfTime = shelfTime;
        batch.volume = articleVolume;
        Stock::ArticleQuants articleQuants;
        articleQuants.push_back(batch);
        std::pair<QString, Stock::ArticleQuants>
                newArticleQuant  (articleId, articleQuants);
        articles.insert(newArticleQuant);
    } else {
        // Mamy jakieś ilości z tym artykułem,
        Stock::ArticleQuants *articleQuants = &(articleIter->second);
        // Szukamy o takiej samej przydatności
        Stock::Batch batch;// = new Batch();
        batch.quant = quant;
        batch.shelfTime = shelfTime;
        batch.volume = articleVolume;
        if (shelfTime > 0 || quant > 0.) {
            Stock::ArticleQuants::iterator findTheSameShelfTime = find(articleQuants->begin(), articleQuants->end(), batch);
            if (findTheSameShelfTime == articleQuants->end()) {
                // Nie znaleźliśmy
                assert (shelfTime);
                assert(quant > 0.);
                articleQuants->push_back(batch);
            } else {
                // Wystarczy zmienić ilość
                findTheSameShelfTime->quant += quant;
                assert ((*findTheSameShelfTime).quant >= 0.0);
            }
        } else {
            // FEFO : shelfTime <= 0 && quant <= 0.
            articleQuants->sort(Batch());
            Stock::ArticleQuants::iterator batchesIt = articleQuants->begin();
            Stock::ArticleQuants::iterator batchesEnd = articleQuants->end();
            while (batchesIt != batchesEnd
                   && quant != 0.) {
                Batch *batch = &(*batchesIt);
                double diff = batch->quant + quant;
                if (diff > 0.) {
                    batch->quant = diff;
                    quant = 0.;
                } else { // Tą pozycję usuwamy i zmniejszamy ilosć
                    batch->quant = 0.;
                    quant = diff;
                }
                batchesIt++;
            }
            if (quant < 0) {
                qFatal("quant %f of article %s in stock %s < 0.",
                       quant, articleIter->first.toStdString().c_str(),
                       getId().toStdString().c_str());
                assert (quant >= 0.);
            }
        }
        cleanUp(articleIter);
    }
}

void Stock::cleanUp(Stock::Articles::iterator &articleIter)
{
    ArticleQuants *list = &(articleIter->second);
    list->remove_if( &Batch::expireOrNothing );
    if (articleIter->second.size () == 0) {
        // Usuwamy z listy artykułów
        articles.erase(articleIter++);
    } else {
        ++articleIter;
    }
}
