#include <QSplitter>
#include <QHBoxLayout>
#include <QtGui/QApplication>

#include <QtSql/QSqlDatabase>
#include <QtSql/QSqlQueryModel>
#include <QtSql/QSqlRecord>
#include <QSqlError>
#include <QDebug>
#include <QPropertyAnimation>

#include <assert.h>

#include "simmainwindow.h"
#include "ui_simmainwindow.h"
#include "fieldview.h"
#include "creatureview.h"

#include <QtGui>

#include "stockinfoview.h"
#include "activitiesinfoview.h"
#include "zoneview.h"
#include "util/random.h"

#include "dialogs/fielddialog.h"

SimMainWindow::SimMainWindow(QWidget *parent, QString database) :
    QMainWindow(parent),
    ui(new Ui::SimMainWindow),
    worldStock ("summaryArticlesQuant", 1000000000000000000, false)
{
    dayNo = 0;
    fieldManager = FieldManager::getManager();
    creatureManager = CreatureManager::getManager();
    speciesManager = SpeciesManager::getManager();
    articleManager = ArticleManager::getManager();
    recipeExecutor = RecipeExecutor::getManager();
    stockManager = StockManager::getManager();
    zoneManager = ZoneManager::getManager();
    naturalPhenomena = NaturalPhenomena::getManager();
    imageManager = ImageManager::getManager();
    paramManager = ParamManager::getManager();

    Messenger *messenger = new Messenger(this);
    creatureManager->setMessenger(messenger);
    stockManager->setMessenger(messenger);
    recipeExecutor->setMessenger(messenger);
    connect (messenger, SIGNAL(creatureBorn(QString)),
             this, SLOT(handleCreatureBorn(QString)),
             Qt::DirectConnection);
    connect (messenger, SIGNAL(creatureActivityDone(QString,QString)),
             this, SLOT(handleCreatureActivity(QString, QString)));
    connect (messenger, SIGNAL(creatureDied(QString)),
             this, SLOT(handleCreatureDie(QString)),
             Qt::DirectConnection);
    connect (messenger, SIGNAL(stockChanged(QString)),
             this, SLOT(handleStockChange(QString)));
    connect (messenger, SIGNAL(articleQuantChange(QString,QString,double)),
             this, SLOT(handleArticleQuantChange(QString,QString,double)),
             Qt::DirectConnection);
    scene = new QGraphicsScene;
    stepTimer = new QTimer(this);
    stepTimer->setInterval(1000);
    stepTimer->setSingleShot(false);

    StockInfoView *worldStockView = new StockInfoView(this);
    connect(this,  SIGNAL(worldStockChanged(const Stock &)),
            worldStockView, SLOT(displayStockInfo(const Stock &)));
    StockInfoView *stockView = new StockInfoView(this);
    connect(this,  SIGNAL(stockChanged(QString)),
            stockView, SLOT(displayStockInfo(QString)));

    animateCreatures = new QCheckBox( tr("Animation on"), this);
    animateCreatures->setChecked(true);
    QString appDir = QApplication::applicationDirPath();
    currentDBName = appDir.append("/").append(database);
    loadWorld (currentDBName);
    populateScene();

    ui->setupUi(this);
    //setWindowState(Qt::WindowMaximized);
    ui->statusBar->showMessage(QString ("World database %1 opened").arg(database), 5000);

    QSplitter *vSplitter = new QSplitter;
    vSplitter->setOrientation(Qt::Vertical);

    view = new View("World view: ");
    view->view()->setScene(scene);

    connect (view->view(), SIGNAL(customContextMenuRequested(QPoint)),
             this, SLOT(handleContextMenuRequest(QPoint)));
    view->view()->setContextMenuPolicy(Qt::CustomContextMenu);

    vSplitter->addWidget(view);


    activitiesView = new ActivitiesInfoView(this);

    QHBoxLayout *layout = new QHBoxLayout (ui->centralWidget);
    QVBoxLayout *vLayout = new QVBoxLayout();
    layout->addWidget(vSplitter, 4);
    layout->addLayout(vLayout);

    vLayout->addWidget(worldStockView);
    vLayout->addWidget(stockView);
    vLayout->addWidget(activitiesView);


    layout->layout();
    mutex = new QMutex;
    locker = new QMutexLocker(mutex);

    startStopAction =  new QPushButton("Start", this);
    startStopAction->setCheckable(true);
    startStopAction->setChecked(false);
    ui->mainToolBar->addWidget(startStopAction);
    connect(startStopAction, SIGNAL(toggled(bool)), this,
            SLOT(setSimulationRun(bool)));

    ui->mainToolBar->addAction("Save", this, SLOT(saveToCurrentDb()));
    connect(stepTimer, SIGNAL(timeout()), this, SLOT(nextDay()));


    QSpinBox *intervalSpinBox = new QSpinBox(this);
    intervalSpinBox->setMaximum(5000);
    intervalSpinBox->setMinimum(50);
    intervalSpinBox->setSingleStep(100);
    intervalSpinBox->setValue(stepTimer->interval());

    connect(intervalSpinBox, SIGNAL(valueChanged(int)), this, SLOT(dayInterval(int)));
    ui->mainToolBar->addWidget(new QLabel(tr("Day duration (ms)")));
    ui->mainToolBar->addWidget(intervalSpinBox);

    QCheckBox *paintCreatureObjVal = new QCheckBox( tr("Show creature value"), this);
    paintCreatureObjVal->setChecked(true);
    ui->mainToolBar->addWidget(paintCreatureObjVal);
    connect (paintCreatureObjVal, SIGNAL(clicked(bool)),
             this, SLOT(setCreaturesObjectiveValuesVisibility(bool)));




    ui->mainToolBar->addWidget(animateCreatures);
  //  connect (animateCreatures, SIGNAL(clicked(bool)),
   //          this, SLOT(setCreaturesObjectiveValuesVisibility(bool)));

}

SimMainWindow::~SimMainWindow()
{
    delete ui;
    delete mutex;
    delete locker;
    delete creatureActivitiesView;

    QMapIterator<QString, CreatureVisual> j (creaturesViews);
    while (j.hasNext()) {
        CreatureVisual v = j.next().value();
        v.animation->stop();
        delete v.animation;
        scene->removeItem(v.view);
        delete v.view;
        if (v.bornAnimation)
            delete v.bornAnimation;
    }
    creaturesViews.clear();

    QListIterator<QGraphicsItem *> i(scene->items());
    while(i.hasNext()) {
        QGraphicsItem *o = i.next();
        scene->removeItem(o);
        delete o;
    }
    delete view;
    delete scene;
    fieldManager->clear();
    creatureManager->clear();
    speciesManager->clear();
    articleManager->clear();
    recipeExecutor->clear();
    stockManager->clear();
    zoneManager->clear();
    imageManager->clear();
    paramManager->clear();
    naturalPhenomena->clear();
}

void SimMainWindow::nextDay()
{
    locker->relock();
    dayNo++;
    bool animate = animateCreatures->isChecked();
    QMapIterator<QString, CreatureVisual> i (creaturesViews);
    while (i.hasNext()) {
        CreatureVisual v = i.next().value();
        v.animation->stop();
        v.animation->setKeyValues(QVariantAnimation::KeyValues());
        assert (v.animation->keyValues().count() == 0);
        if (animate) {
            v.animation->setDuration(stepTimer->interval());
            v.animation->setKeyValueAt(0, v.view->pos());
            if (i.key() ==lastSelectedCreatureId ) {
                creatureActivitiesView->setPos(v.view->pos());

            }
        }
    }
    // Wyszukaj wszystkie osobniki , które znajdują się w świecie i niech każdy
    // wykonuje swoje czynności

    // TODO : Podzielić zwracany zbiór na podzbiory i wykonywać je w wątku
    CreatureManager::Cache::const_iterator creatureIt = creatureManager->getCache().begin();
    CreatureManager::Cache::const_iterator creatureEnd = creatureManager->getCache().end();
    while (creatureIt != creatureEnd) {
         creatureManager->dayActivity(creatureIt->first);
         creatureIt++;
    }

    if (animate) {
        i.toFront();
        while (i.hasNext()) {
            CreatureVisual v = i.next().value();
            if (v.animation->keyValues().count() > 1)
                v.animation->start();
        }
    }
    locker->unlock();

    // Na wszystkich polach wszystkie gatunki mnożą się i umierają
    ZoneManager::Cache::const_iterator zoneIt = zoneManager->getCache().begin();
    ZoneManager::Cache::const_iterator zoneEnd = zoneManager->getCache().end();
    while (zoneIt != zoneEnd) {
        zoneManager->reproduction(zoneIt->second);
        zoneManager->evolutionaryPressure(zoneIt->second);
        zoneIt++;
    }

    // Wszystkie zjawiska doświadczają upływu czasu
    naturalPhenomena->passageOfTime(dayNo);

    // Zaktualizuj wszystkie stany
    StockManager::Cache::const_iterator stockIt = stockManager->getCache().begin();
    StockManager::Cache::const_iterator stockEnd = stockManager->getCache().end();
    while (stockIt != stockEnd) {
        stockManager->dayPassed(stockIt->first);
        stockIt++;
    }

    ui->statusBar->showMessage(QString ("Day %1 passed").arg(dayNo), 5000);
    emit stockChanged (lastSelectedStockId);
}


void SimMainWindow::handleCreatureAddOrChange(QString creatureId)
{
    QColor greenColor("#666666");
    // Dodaj reprezentację osobników
    const Creature *creature = creatureManager->find(creatureId);

    QGraphicsObject *item = new CreatureView(creature->getId(), greenColor);
    item->setPos(QPointF(creature->getCoordinates().x() * FieldView::FIELD_SIZE,
                         creature->getCoordinates().y() * FieldView::FIELD_SIZE));
    scene->addItem(item);
    connect(item, SIGNAL(creatureSelected(QString)),
            this, SLOT(handleCreatureSelection(QString)));
    CreatureVisual visual;
    visual.view = item;
    visual.animation = new QPropertyAnimation(item, "pos");
    creaturesViews.insert(creatureId, visual);

    if (animateCreatures->isChecked() ) {
            visual.bornAnimation = new QParallelAnimationGroup (item);
            QPropertyAnimation * bornAnimationScale = new QPropertyAnimation(item, "scale", visual.bornAnimation);
            bornAnimationScale->setDuration(stepTimer->interval());
            bornAnimationScale->setKeyValueAt(0, 1);
            bornAnimationScale->setKeyValueAt(0.5, 0.001);
            bornAnimationScale->setKeyValueAt(1, 1.0);
            bornAnimationScale->setEasingCurve(QEasingCurve::OutElastic);

            QPropertyAnimation * bornAnimationRotation = new QPropertyAnimation(item, "rotation", visual.bornAnimation);
            bornAnimationRotation->setDuration(stepTimer->interval() / 2);
            bornAnimationRotation->setKeyValueAt(0, 90);
            bornAnimationRotation->setKeyValueAt(1, 360);
            visual.bornAnimation->addAnimation(bornAnimationRotation);
            visual.bornAnimation->addAnimation(bornAnimationScale);
            visual.bornAnimation->start();
    }
    qDebug() << "creature "<< creatureId << " born and added to scene";
}

void SimMainWindow::handleCreatureBorn(QString creatureId)
{
    locker->relock();
    //const Creature *creature = creatureManager->find(creatureId);
    handleCreatureAddOrChange(creatureId);
    locker->unlock();
}

void SimMainWindow::handleFieldAddOrChange(QString fieldId)
{
    const Field *field = fieldManager->find(fieldId);
    if (field) {
        ObjectsViews::iterator it = fieldsViews.find(fieldId);
        QGraphicsObject *item = NULL;
        if (it != fieldsViews.end())
            item = it->second;
        if (item == NULL) {
            item = new FieldView(field->getId(), QColor::fromRgb(100, 100, 200),
                                 field->getResourcesId());

            int xPos = field->getCoordinates().x () * FieldView::FIELD_SIZE;
            int yPos = field->getCoordinates().y () * FieldView::FIELD_SIZE;
            item->setPos(xPos, yPos);
            scene->addItem(item);

            qDebug() << "FIELD " << fieldId << "added" << scene->items().count();

            std::pair<QString, QGraphicsObject *> fView (field->getId(), item);
            fieldsViews.insert(fView);
            connect(item, SIGNAL(fieldSelected(QString)),
                    this, SLOT(handleFieldSelection(QString)));
        } else {
            qWarning() << "TODO : handle change ";
        }
    }
}

void SimMainWindow::handleCreatureActivity(QString creatureId, QString activityId)
{
    Q_UNUSED(activityId);
    // Dodaj reprezentację osobników
    const Creature *creature = creatureManager->find(creatureId);
    CreaturesVisuals::iterator f = creaturesViews.find(creatureId);
    CreatureVisual v = f.value();
    QPoint newPos (creature->getCoordinates().x() * FieldView::FIELD_SIZE,
                   creature->getCoordinates().y() * FieldView::FIELD_SIZE);
    const Species *species = speciesManager->find(creature->getSpeciesId());
    // Pomniejszamy o jeden bo dodaliśmy na starcie początkowy krok.
    qreal steps = v.animation->keyValues().count();
    qreal step = steps / species->getActivitiesCount();
    v.animation->setKeyValueAt(step, newPos);


}

void SimMainWindow::handleCreatureDie(QString creatureId)
{
    locker->relock();
    qDebug() << "handleCreatureDie " << creatureId;
    CreaturesVisuals::iterator f = creaturesViews.find(creatureId);
    QGraphicsObject *item = f.value().view;
    QPropertyAnimation *anim = f.value().animation;
    anim->stop();
    scene->removeItem(f.value().view);
    delete anim;
    delete item;
    creaturesViews.erase(f);

    if (lastSelectedCreatureId == creatureId) {
        lastSelectedCreatureId = QString::null;
        activitiesView->displayActivitiesInfo(lastSelectedCreatureId);
        if (creatureId == creatureActivitiesView->getElementId()) {
            creatureActivitiesView->setElementId(QString::null);
        }
    }
    locker->unlock();
}

void SimMainWindow::handleStockChange(QString stockId)
{
    if (stockId == lastSelectedStockId) {
        // Zmienił się zasób aktualnie zaznaczonego obiektu
         emit stockChanged (stockId);
    }
    // Poszukajmy pola lub marketu związanego z zasobem i odświeżmy widok
    QGraphicsObject *item = NULL;
    ObjectsViews::iterator f = fieldsViews.find(stockId);
    if (f != fieldsViews.end()) {
        item = f->second;
    }
    if (item) {
        item->update();
    }

}

void SimMainWindow::handleArticleQuantChange(QString stockId, QString articleId, double delta)
{
    Q_UNUSED(stockId);
   // qDebug() << "handleArticleQuantChange " << stockId << ", " << articleId << "," << delta;
    const Stock *stock = stockManager->find(stockId);
    if ( stock && stock->getChangeShelfTime() ) {
        // Uwzględniamy zapasy, które są zależne od czasu
        if (delta > 0.) {
            worldStock.push(articleId, delta);
        } else {
            worldStock.pop(articleId, -delta);
        }
        emit worldStockChanged(worldStock);
    }
}

void SimMainWindow::handleCreatureSelection(QString creatureId)
{
    const Creature *creature = creatureManager->find(creatureId);
    if (creature) {
        lastSelectedStockId = creature->getBackpackId();
        lastSelectedCreatureId = creatureId;
        emit stockChanged (lastSelectedStockId);
        activitiesView->displayActivitiesInfo(creatureId);
        creatureActivitiesView->setElementId(creatureId);
        QPoint newPos (creature->getCoordinates().x() * FieldView::FIELD_SIZE,
                       creature->getCoordinates().y() * FieldView::FIELD_SIZE);
        if (lastSelectedCreatureId == creatureId) {
            creatureActivitiesView->setPos(newPos);
        }
    } else {
        qDebug() << " creature " << creatureId << " not found " ;
    }
}

void SimMainWindow::handleFieldSelection(QString fieldId)
{
    Q_UNUSED(fieldId);
    const Field *field = fieldManager->find(fieldId);
    if (field) {
        lastSelectedStockId = field->getResourcesId();
        emit stockChanged (lastSelectedStockId);
    }
}

void SimMainWindow::handleContextMenuRequest(QPoint point) {

    QMenu menu(this);

    menu.addAction(tr("Add field"), this, SLOT(modifyFieldAtPoint()));

    QPointF map = view->view()->mapToScene(point);

    int x = map.x();
    int y = map.y();
    if (x < 0) {
        x -= FieldView::FIELD_SIZE;
    }
    if (y < 0) {
        y -= FieldView::FIELD_SIZE;
    }
    worldCoordinates.setX(x / FieldView::FIELD_SIZE);
    worldCoordinates.setY(y / FieldView::FIELD_SIZE);

    menu.exec(QCursor::pos());

}

void SimMainWindow::modifyFieldAtPoint()
{
    qDebug() << "Adding new field";
    Field field(QString("F%1x%2").arg(worldCoordinates.x()).arg(worldCoordinates.y()),
                worldCoordinates.x(), worldCoordinates.y());
    FieldDialog *dialog = new FieldDialog(this);
    dialog->setField(field);
    if (dialog->exec() == QDialog::Accepted) {
        handleFieldAddOrChange(dialog->fieldId());
    }
}

void SimMainWindow::setSimulationRun(bool run)
{
    if (run) {
        stepTimer->start();
        startStopAction->setText(tr("Stop"));
    } else {
        startStopAction->setText(tr("Start"));
        stepTimer->stop();
    }
}

void SimMainWindow::setCreaturesObjectiveValuesVisibility(bool show)
{
    CreatureView::paintObjectiveValue = show;
    scene->update();
}

void SimMainWindow::dayInterval(int dayInMilisecond)
{
    stepTimer->setInterval(dayInMilisecond);
}

void SimMainWindow::saveToCurrentDb()
{
    locker->relock();
    // Aktuallizacja parametró w
    QVariant nextId((int) creatureManager->getNextObjectId ());
    Parameter nextObjId ("TECH/CREATURES/NEXTOBJECTID",  nextId);
    paramManager->merge(nextObjId);

    QVariant dayNoValue(dayNo);
    Parameter dayNoParam ("TECH/APP/DAYNO",  dayNoValue);
    paramManager->merge(dayNoParam);

    qDebug() << "saveToCurrentDb";
    QSqlDatabase::database().transaction();
    if ( imageManager->saveAll()
         && articleManager->saveAll()
         && recipeExecutor->saveAll()
         && stockManager->saveAll()
         && speciesManager->saveAll()
         && zoneManager->saveAll()
         && naturalPhenomena->saveAll()
         && fieldManager->saveAll()
         && creatureManager->saveAll()
         && paramManager->saveAll()
         && creatureManager->applyRemove()
         && fieldManager->applyRemove()
         && zoneManager->applyRemove()
         && speciesManager->applyRemove()
         && stockManager->applyRemove()
         && recipeExecutor->applyRemove()
         && articleManager->applyRemove()
         && paramManager->applyRemove()
         && naturalPhenomena->applyRemove()
         )
        QSqlDatabase::database().commit();
    else {
        QSqlDatabase::database().rollback();
        qDebug( ) <<QSqlDatabase::database().lastError().driverText();
        qFatal("%s", " error during save ");
    }
    locker->unlock();
}

void SimMainWindow::populateScene()
{

    // Na wszystkich polach wszystkie gatunki mnożą się i umierają
    ZoneManager::Cache::const_iterator zoneIt = zoneManager->getCache().begin();
    ZoneManager::Cache::const_iterator zoneEnd = zoneManager->getCache().end();
    while (zoneIt != zoneEnd) {
        const Zone *zone = &(zoneIt->second);
        QGraphicsObject *zoneView = new ZoneView (zone->getId(), QColor::fromRgb(200, 200, 200),
                                                zone->getArea().width() *FieldView::FIELD_SIZE,
                                                zone->getArea().height()*FieldView::FIELD_SIZE,
                                                NULL);
        zoneView->setPos(zone->getArea().x()*FieldView::FIELD_SIZE,
                         zone->getArea().y()*FieldView::FIELD_SIZE);
        scene->addItem(zoneView);
        qDebug() << "ZONE " << zone->getId() << "added" << scene->items().count();

        zoneIt++;
    }

    FieldManager::Cache::const_iterator fieldIt = fieldManager->getCache().begin();
    FieldManager::Cache::const_iterator fieldEnd = fieldManager->getCache().end();
    while (fieldIt != fieldEnd) {
        handleFieldAddOrChange(fieldIt->first);
        fieldIt++;
    }

    CreatureManager::Cache::const_iterator creatureIt = creatureManager->getCache().begin();
    CreatureManager::Cache::const_iterator creatureEnd = creatureManager->getCache().end();
    while (creatureIt != creatureEnd) {
        handleCreatureAddOrChange(creatureIt->first);
        creatureIt++;
    }
    creatureActivitiesView = new ActivitiesView(QString::null, QColor(Qt::red));
    scene->addItem(creatureActivitiesView);

    qDebug() << " creatureActivitiesView added" << scene->items().count();

}

void SimMainWindow::loadWorld(QString databaseFile)
{
    QSqlDatabase db = QSqlDatabase::addDatabase("QSQLITE");
    db.setDatabaseName(databaseFile);
    if ( db.isValid() && db.open()) {
        if ( paramManager->load()
             && imageManager->load()
             && articleManager->load()
             && recipeExecutor->load()
             && stockManager->load()
             && speciesManager->load()
             && zoneManager->load()
             && naturalPhenomena->load()
             && fieldManager->load()
             && creatureManager->load() ) {
            qDebug() << " data loaded" ;

            const Parameter *nextId = paramManager->find("TECH/CREATURES/NEXTOBJECTID");
            if (nextId != NULL) {
                creatureManager->setNextObjectId (nextId->value().toUInt());
            }
            const Parameter *dayNoParam = paramManager->find("TECH/APP/DAYNO");
            if (dayNoParam != NULL) {
                dayNo = dayNoParam->value().toInt();
            }
        }
        else
            qFatal ("%s", "Cannot load data");
    }
    else {
        qFatal("%s", QString ("cannot open database : %1").arg(databaseFile).toStdString().c_str() );
    }

}
