#include "recipeexecutiontest.h"
#include "model/creature.h"
#include "controller/articlemanager.h"

RecipeExecutionTest::RecipeExecutionTest() :
    ExampleData()
{
}

void RecipeExecutionTest::setUp()
{
    ExampleData::setUp();

    recipe = new Recipe("common", Container::CREATURE);

    /*
    Osobnik będzie posiadał plecak, o początkowej pustej zawartości
    i będzie miał namiary na sklep z pustymi pułkami
    */
    creature = new Creature ("creature");


    creatureStock = new Stock("creatureStock", 100);
    marketStock =   new Stock("marketStock", 1000);;
    fieldStock =    new Stock("fieldStock", 10000);

    std::pair<Container::Type, Stock *> stocksElem;

    stocksElem.first = Container::CREATURE;
    stocksElem.second = creatureStock;
    stocks.insert(stocksElem);
    stocksElem.first = Container::MARKET;
    stocksElem.second = marketStock;
    stocks.insert(stocksElem);
    stocksElem.first = Container::FIELD;
    stocksElem.second = fieldStock;
    stocks.insert(stocksElem);

    recipeExecutor->merge(*coinageRecipe);
    zoneManager->clear();
}

void RecipeExecutionTest::tearDown()
{
    ExampleData::tearDown();
    delete recipe;
    delete creature;
    delete creatureStock;
    delete marketStock;
    delete fieldStock;
    stocks.clear();
    //stocks.assign(0, NULL);
}

void RecipeExecutionTest::buyTest() {

    /*
      Surowiec:
        skąd pobieramy: zapasy inicjatora
        typ: waluta
        ilość: 1.00
      */
    currencySource = Container::CREATURE;
    breadPrice = 1.99;

    /*
       Produkty:
         [ gdzie składujemy: zapasy inicjatora
           typ: chleb
           ilość: 0.5 ],
         [ gdzie składujemy: zapasy sklepu
           typ: waluta
           ilość: 1.99 ]
     */
    breadDestination = Container::CREATURE;
    breadQuantity = 0.5;
    currencyDestination = Container::MARKET;

    /*
      Surowiec:
        skąd pobieramy: zapasy inicjatora
        typ: waluta
        ilość: 1.00
    */
    recipe->addResource(currencySource, coin->getId(), breadPrice);
    /*
       Produkty:
         [ gdzie składujemy: zapasy inicjatora
           typ: chleb
           ilość: 0.5 ],
         [ gdzie składujemy: zapasy sklepu
           typ: waluta
           ilość: 1.99 ]
     */
    recipe->addProduct(breadDestination, bread->getId(), breadQuantity);
    recipe->addProduct(currencyDestination, coin->getId(), breadPrice);

    // Osobnik nie ma pieniędzy, więc wykonanie receptury nie powinno się udać
    bool executed = recipeExecutor->execute(*recipe, stocks);
    CPPUNIT_ASSERT(!executed);
    /*
    Zmieniamy zawartość plecaka osobnika
        TYP PRODUKTU    | ILOŚĆ
        waluta          |   3.0
        chleb           |   0.2
    */
    double changerWallet = 3.0;
    double changerBreadQuant = 0.2;
    double marketWallet = 0.0;

    creatureStock->push(*coin, changerWallet);
    creatureStock->push(*bread, changerBreadQuant);

    // Osobnik ma już odpowiednią ilość kasy więc powinno udać się wykonać recepturę kupowania ...
    executed = recipeExecutor->execute(*recipe, stocks);
    // oraz plecak osobnika oraz stany sklepu powinny się zmienić.
    CPPUNIT_ASSERT(executed);
    // Zmniejszyła się wartość portfela
    double newChangerCurrencyQuant = creatureStock->summaryQuant(coin->getId());
    CPPUNIT_ASSERT(newChangerCurrencyQuant == changerWallet - breadPrice);
    // Osobnik ma więcej chleba
    double newChangerBredQuant = creatureStock->summaryQuant(bread->getId());
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Ilość chleba w plecaku powinna się zwiększyć",
            changerBreadQuant + breadQuantity, newChangerBredQuant );
    // W sklepie powinno być więcej waluty.
    double newMarketWallet = marketStock->summaryQuant(coin->getId());
    CPPUNIT_ASSERT(newMarketWallet == marketWallet + breadPrice);
}

void RecipeExecutionTest::moveTest()
{
    /*
      Receptura wymaga posiadania od osobnika energi na ruch,
      wykorzystują ją i zmienia położenie osobnika
    */
    // Zawsze poruszamy się w jakimś świecie
    Zone f ("f");
    f.setArea(Area (1, 1, 5, 5));
    zoneManager->merge(f);
    recipe->addResource(Container::CREATURE, energy->getId(), 1);
    recipe->setPositionEffect(-1, 1);
    bool executed = recipeExecutor->execute(*recipe, stocks, *creature);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Nie można poruszać się nie posiadając energii",
                                 false, executed);

    CPPUNIT_ASSERT_EQUAL_MESSAGE("Nowe wspołrzędne osognika po ruchu 1",
                                 (int) 0, creature->getCoordinates().first);

    // Zasilamy osobnika energią
    creatureStock->push(*energy, 10);
    executed = recipeExecutor->execute(*recipe, stocks, *creature);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Energetyczny osobnik powinien sie ruszyć",
                                 true, executed);

    CPPUNIT_ASSERT_EQUAL_MESSAGE("Nowe wspołrzędne x osognika po ruchu 1 "
                                 "(przeszliśmy na drugą strone i jeszcze krok "
                                 "(bo startowalismy spoza pola))",
                                 (int) 3, creature->getCoordinates().first);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Nowe wspołrzędne y osognika po ruchu 1 ",
                                 (int) 1, creature->getCoordinates().second);


    CPPUNIT_ASSERT_EQUAL_MESSAGE("Energia po wykonaniu ruchu powinna być pomniejszona",
                                 (double) 9, creatureStock->summaryQuant(energy->getId()));
    recipe->setPositionEffect(-1, 10);
    executed = recipeExecutor->execute(*recipe, stocks, *creature);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Energetyczny osobnik powinien sie ruszyć",
                                 true, executed);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Energia po wykonaniu ruchu powinna być pomniejszona",
                                 (double) 8, creatureStock->summaryQuant(energy->getId()));


    CPPUNIT_ASSERT_EQUAL_MESSAGE("Nowe wspołrzędne x osognika po ruchu 1 "
                                 "(przeszliśmy na drugą strone i jeszcze krok "
                                 "(bo startowalismy spoza pola))",
                                 (int) 2, creature->getCoordinates().first);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Wspólrzędna nie powinna się zmienić",
                                 (int) 1, creature->getCoordinates().second);


}

void RecipeExecutionTest::growTest()
{
    // Tworzymy recepturę
    recipe->addResource(Container::FIELD, corn->getId(), 3);
    recipe->addResource(Container::FIELD, soil->getId(), 1);
    recipe->addResource(Container::FIELD, water->getId(), 1);
    recipe->addProduct(Container::FIELD, grain->getId(), 5);

    bool executed = recipeExecutor->execute(*recipe, stocks);
    CPPUNIT_ASSERT_MESSAGE("Nic nie może urosnąć jeżelli brak na stanie któregokolwiek składnika",
                           executed == false);
    // Dodajemy po kolei składniki
    fieldStock->push(*soil);
    executed = recipeExecutor->execute(*recipe, stocks);
    CPPUNIT_ASSERT(executed == false);

    fieldStock->push(*water);
    executed = recipeExecutor->execute(*recipe, stocks);
    CPPUNIT_ASSERT(executed == false);

    fieldStock->push(*corn);
    executed = recipeExecutor->execute(*recipe, stocks);
    CPPUNIT_ASSERT(executed == false);

    fieldStock->push(*corn, 2.0);
    executed = recipeExecutor->execute(*recipe, stocks);
    CPPUNIT_ASSERT_MESSAGE("Mamy wszystkie składniki więc powinniśmy móc urosnąć...",
                           executed);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Na polu pojawiły się łany zboża",
                                 (double) 5, fieldStock->summaryQuant(grain->getId()));
}

void RecipeExecutionTest::takeTest()
{
    // Tworzymy recepturę
    recipe->addResource(Container::FIELD, corn->getId(), 3);
    recipe->addProduct(Container::CREATURE, corn->getId(), 3);
    recipe->addResource(Container::CREATURE, energy->getId(), 1);
    bool executed = recipeExecutor->execute(*recipe, stocks);
    CPPUNIT_ASSERT_MESSAGE("Nie można pobrać bo osobnik nie ma energii",
                           executed == false);
    // Dodajemy po kolei składniki
    creatureStock->push(*energy);
    executed = recipeExecutor->execute(*recipe, stocks);
    CPPUNIT_ASSERT_MESSAGE("Nie można pobrać bo na polu nie ma ziarna",
                           executed == false);
    fieldStock->push(*corn, 3);
    executed = recipeExecutor->execute(*recipe, stocks);
    CPPUNIT_ASSERT_MESSAGE("Udało się pobrać, ", executed == true);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Z pola znikło ziarno",
                                 (double) 0, fieldStock->summaryQuant(corn->getId()));

    CPPUNIT_ASSERT_EQUAL_MESSAGE("Osonik nie ma juz energii",
                                 (double) 0, creatureStock->summaryQuant(energy->getId()));

    CPPUNIT_ASSERT_EQUAL_MESSAGE("Ale ma za to ziarno",
                                 (double) 3, creatureStock->summaryQuant(corn->getId()));
}

void RecipeExecutionTest::makeCoinsTest()
{
    bool executed = recipeExecutor->execute(*coinageRecipe, stocks);
    CPPUNIT_ASSERT_MESSAGE("Nie można produkować monet bo brak surowców",
                           executed == false);
    fieldStock->push(*iron, 1);
    fieldStock->push(*wood, 1);
    fieldStock->push(*fire, 1);
    marketStock->push(*coin, 1);
    executed = recipeExecutor->execute(*coinageRecipe, stocks);
    CPPUNIT_ASSERT_MESSAGE("Mamy wszystkie surowce, więc produkcja Ok",
                           executed == true);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("W markecie pojawił się dodatkowo 10 monet",
                                 (double) 11, marketStock->summaryQuant(coin->getId()));
}



