#include "stocktest.h"
#include "articlemanager.h"

StockTest::StockTest() :
    CppUnit::TestCase()
{
}

void StockTest::setUp()
{
}

void StockTest::tearDown()
{
}

void StockTest::stockTest()
{
    // Nowo stworzony stos jest pusty i ma zerową pojemność
    Stock *stock = new Stock ("stock");
    CPPUNIT_ASSERT(stock->isEmpty());
    CPPUNIT_ASSERT(stock->availableCapacity() == 0.0);
    CPPUNIT_ASSERT(stock->volumeUsed() == 0.0);

    // Po ustawieniu pojemności ma ona odpowiednią wartość
    double stockCapacity = 100.0;
    stock->setCapacity (stockCapacity);
    CPPUNIT_ASSERT(stock->availableCapacity() == stockCapacity);
    QString articleId ("beer");

    // Po dodaniu artykułu o objętość 0.5 pojemność kontenera zmniejsza się
    double bottleVolume = 0.5;
    Article beer60 (articleId, bottleVolume, 60);
    double oneBottle = 1.0;
    stock->push(beer60, oneBottle, beer60.getShelfTime());
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Wykorzystana pojemność zwiększyła się o nieprawidłową ilość",
                                 beer60.volume(oneBottle),
                                 stock->volumeUsed());

    CPPUNIT_ASSERT_EQUAL_MESSAGE("Pojemność nie została zmniejszona",
                                 stockCapacity - beer60.volume(oneBottle),
                                 stock->availableCapacity());

    // Po dodaniu piwa o innej dacie ważności łączna ich liczba będzie = 2
    Article beer30 (articleId, bottleVolume, 30);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Nieprawidłowa liczba partii towaru",
                                 1, stock->batchesCount(articleId));

    stock->push(beer30, oneBottle);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Nieprawidłowa liczba partii towaru",
                                 2, stock->batchesCount(articleId) );

    stock->push(beer30, oneBottle);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Nieprawidłowa łączna ilość towaru",
                                 3 * oneBottle,
                                 stock->summaryQuant(articleId));

    stock->pop(beer30, oneBottle);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Nieprawidłowa łączna ilość towaru",
                                 2 * oneBottle,
                                 stock->summaryQuant(articleId));

    stock->pop(beer30, oneBottle / 2.0);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Nieprawidłowa łączna ilość towaru",
                                 (double) 1.5 * oneBottle,
                                 stock->summaryQuant(articleId));

    // Jeszcze kilka sprawdzeń, po dodaniu innego towaru
    Article a2 ("a2", 2.2, 33);
    double f = stock->volumeUsed();
    stock->push(a2, 1.4);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Nieprawidłowa objętość",
                                 (double) (2.2 * 1.4) + f,
                                 stock->volumeUsed() )  ;

    CPPUNIT_ASSERT_EQUAL_MESSAGE("Nieprawidłowa liczba partii towaru",
                                 1, stock->batchesCount("a2"));

    delete stock;
}



void StockTest::dayPassedTest()
{
    // Nowo stworzony stos jest pusty i ma zerową pojemność
    Stock *stock = new Stock ("stock", 100.0);
    Article a1 ("a1", 1.1, 11);
    Article a2 ("a2", 2.2, 22);
    Article a3 ("a3", 3.3, 33);

    ArticleManager::getManager()->merge(a1);
    ArticleManager::getManager()->merge(a2);
    ArticleManager::getManager()->merge(a3);

    stock->push(a1, 10, 11);

    stock->push(a2, 5, 11);
    stock->push(a2, 2, 22);

    stock->push(a3, 10, 12);
    stock->push(a3, 10, 22);
    stock->push(a3, 1, 33);

    int days = 1;
    for (; days <= 10; days++) {
        stock->dayPassed();
    }

    double used = stock->volumeUsed();
    double expected = a1.volume(10) + a2.volume(2 + 5) + a3.volume(1 + 10 + 10);
    double accuracy = 0.00001;

    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("Na początku suma objętości wszystkich artykułów daje objętość stosu",
                                         expected,
                                         used,
                                         accuracy);


    stock->dayPassed();
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("Jedenastego zniło 10 a1, 5 a2",
                                         a2.volume(2) + a3.volume(1 + 10 + 10),
                                         stock->volumeUsed(),
                                         accuracy);
    stock->dayPassed();
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("Dwunastego znikło 10 a3",
                                         a2.volume(2) + a3.volume(1 + 10),
                                         stock->volumeUsed(),
                                         accuracy);
    for (days= 13; days <= 22; days++) {
        stock->dayPassed();
    }

    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("Dwudziestego drugiego został jedynie a3",
                                         a3.volume(1),
                                         stock->volumeUsed(),
                                         accuracy);
    for (days= 23; days <= 33; days++) {
        stock->dayPassed();
    }
    CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE("33 dnia pusty zasób",
                                         (double) 0,
                                         stock->volumeUsed(),
                                         accuracy);

    delete stock;
}


void StockTest::fefoTest()
{

    Stock stock ("source", 100.0);
    Article a ("a", 1.1, 11);

    ArticleManager::getManager()->merge(a);
    stock.push(a, 10, 11);
    stock.push(a, 2, 5);
    stock.push(a, 1, 2);

    CPPUNIT_ASSERT_EQUAL_MESSAGE("Na stosie mamy sumę dodanych ilości artykułów",
                                 (double) 13,
                                 stock.summaryQuant(a.getId()));
    stock.pop(a, 1);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Na stosie mamy o jeden mniej",
                                 (double) 12,
                                 stock.summaryQuant(a.getId()));

    CPPUNIT_ASSERT_EQUAL_MESSAGE("Zabraliśmy z artykułu, któr pierwszy straci ważność",
                                 (double) 0,
                                 stock.summaryQuant(a.getId(), 2));
    stock.pop(a, 7);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Na stosie mamy o 7 mniej",
                                 (double) 5,
                                 stock.summaryQuant(a.getId()));

    CPPUNIT_ASSERT_EQUAL_MESSAGE("Nie zostało nic o przydatności 5",
                                 (double) 0,
                                 stock.summaryQuant(a.getId(), 5));
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Zostało 5 sztuk z przydatnością 11",
                                 (double) 5,
                                 stock.summaryQuant(a.getId(), 11));
    stock.push(a, 2, 33);
    stock.pop(a, 1);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Mamy 6 sztuk",
                                 (double) 6,
                                 stock.summaryQuant(a.getId()));

    stock.pop(a, 2);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Mamy 2 sztuk z przydatnością 11",
                                 (double) 2,
                                 stock.summaryQuant(a.getId(), 11));
    stock.pop(a, 4);
    CPPUNIT_ASSERT_EQUAL_MESSAGE("Mamy 0 sztuk",
                             (double) 0,
                                 stock.summaryQuant(a.getId()));
}

void StockTest::zeroShelfTimeAffectionTest()
{

    Stock stock ("source", 100.0);
    Article a ("a", 1.1, 11);
    ArticleManager::getManager()->merge(a);
    stock.push(a, 10, 11);

    int days;
    for (days= 1; days <= 11; days++) {
        stock.dayPassed();
    }

    CPPUNIT_ASSERT_EQUAL_MESSAGE("Upływ czasu domyślnie wpływa na towary "
                                 "i znikają one po upływie terminu",
                             (double) 0,
                                 stock.summaryQuant(a.getId()));

    Stock stockBeyondTime ("source", 100.0, false);
    stockBeyondTime.push(a, 10, 11);
    for (days= 1; days <= 11; days++) {
        stockBeyondTime.dayPassed();
    }

    CPPUNIT_ASSERT_EQUAL_MESSAGE("Upływ czasu domyślnie wpływa na towary "
                                 "i znikają one po upływie terminu",
                                 (double) 10,
                                 stockBeyondTime.summaryQuant(a.getId()));

}
